package rsecc8;
import java.util.Arrays;
import java.util.Random;

// rsecc8.java - RS ecc GF(256) test program
//             - 2017JUL14 18:50

public class rsecc8 {

    // GF(256) 30 x {poly, alpha}
    //  0x11b,0x03,0x11d,0x02,0x12b,0x02,0x12d,0x02,
    //  0x139,0x03,0x13f,0x03,0x14d,0x02,0x15f,0x02,
    //  0x163,0x02,0x165,0x02,0x169,0x02,0x171,0x02,
    //  0x177,0x03,0x17b,0x0b,0x187,0x02,0x18b,0x0b,
    //  0x18d,0x02,0x19f,0x03,0x1a3,0x03,0x1a9,0x02,
    //  0x1b1,0x07,0x1bd,0x07,0x1c3,0x02,0x1cf,0x02,
    //  0x1d7,0x07,0x1dd,0x07,0x1e7,0x02,0x1f3,0x0d,
    //  0x1f5,0x02,0x1f9,0x03

    static byte[] abExp = new byte[512];
    static byte[] abLog = new byte[256];

    static final int  POLY = 0x11d;     // GF(256) polynomial
    static final byte ALPHA = 0x02;     // alpha for GF(256)
    static final byte FCR = 0x01;       // first consecutive root (1 or ALPHA)

    static final int  NDATA = 80;       // number of data bytes
    static final int  NPARITY = 80;     // number of parity bytes
    static final int  NP1 = NPARITY+1;
    static final int  NP2 = NPARITY+2;
    static final int  MAXERR = NPARITY/2;       // max errors correctable
    static final int  NCDWRD = NDATA+NPARITY;   // number of codeword bytes

    static byte[] abGenRoots = new byte[NPARITY];   // generator poly roots
    static byte[] abGenPoly = new byte[NP1];        // generator poly
    static byte[] abCdWrd = new byte[NCDWRD];       // codeword
    static byte[] abCdWrdx = new byte[NCDWRD];      // copy of code word
    static byte[] abParities = new byte[NPARITY];   // parities
    static byte[] abSyndromes = new byte[NPARITY];  // syndromes
    static byte[] abLambdaR = new byte[NP1];        // error locator polynomial
    static byte[] abErrLoc = new byte[NPARITY];     // error locators
    static byte[] abOmega = new byte[NP1];          // error evaluator polynomial
    static byte[] abLambda = new byte[NP1];         // error 1/locator polynomial
    static byte[] abDLambda = new byte[NP1];        // "derivative" of Lambda
    static byte[] abForney = new byte[NP1];         // error values
    // Lambda  has roots 1/(ALPHA^location)
    // LambdaR has roots ALPHA^location == Lambda reversed
    // Sigma is an alternate name for Lambda in some articles

    // extended Euclid data
    static byte[] abE0 = new byte[NP2]; // R[i-2] | A[i-1]
    static byte[] abE1 = new byte[NP2]; // R[i-1] | A[i-2]
    static byte[] abET;                 // temp reference
    static int iE0;                     // index for E0
    static int iE1;                     // index for E1
    static int iET;                     // temp index
    static int nDLambda;                // size of abDLambda
    static int nErr;                    // number of errors
    static int nErrMin = MAXERR;        // min number of errors tested
    static int nErrMax = 0;             // max number of errors tested

    //------------------------------------------------------------------//
    //      GFMpy0(b0,b1)           b0*b1       using low level math    //
    //------------------------------------------------------------------//
    public static byte GFMpy0(byte b0, byte b1)
    {
    int i;
    int product;
        product = 0;
        for(i = 0; i < 8; i++){
            product <<= 1;
            if((product & 0x100) != 0)
                product ^= POLY;
            if((b0 & 0x80) != 0){
                product ^= b1;}
            b0 <<= 1;}
        return((byte)product);
    }

    //------------------------------------------------------------------//
    //      GFAdd(b0, b1)           b0+b1                               //
    //------------------------------------------------------------------//
    static byte GFAdd(byte b0, byte b1)
    {
        return((byte)(b0^b1));
    }

    //------------------------------------------------------------------//
    //      GFSub(b0, b1)           b0-b1                               //
    //------------------------------------------------------------------//
    static byte GFSub(byte b0, byte b1)
    {
        return((byte)(b0^b1));
    }

    //------------------------------------------------------------------//
    //      GFMpy(b0, b1)           b0*b1       using logs              //
    //------------------------------------------------------------------//
    static byte GFMpy(byte b0, byte b1)
    {
        if(b0 == 0 || b1 == 0)
            return((byte)0);
        return(abExp[(abLog[b0&0xff]&0xff)+(abLog[b1&0xff]&0xff)]);
    }

    //------------------------------------------------------------------//
    //      GFDiv(b0, b1)           b0/b1                               //
    //------------------------------------------------------------------//
    static byte GFDiv(byte b0, byte b1)
    {
        if(b1 == 0){
            System.out.println("divide by zero");
            return((byte)0);}
        if(b0 == 0)
            return((byte)0);
        return(abExp[(abLog[b0&0xff]&0xff)-(abLog[b1&0xff]&0xff)+0xff]);
    }
    //------------------------------------------------------------------//
    //      GFPow(b0, b1)           b0^b1                               //
    //------------------------------------------------------------------//
    static byte GFPow(byte b0, byte b1)
    {
    byte b;
        b = 1;
        while(b1 != 0){
            if((b1&1) != 0)
                b = GFMpy(b, b0);
            b0 = GFMpy(b0, b0);
            b1 = (byte)((b1&0xff)>>1);}
        return(b);
    }

    //------------------------------------------------------------------//
    //      Root2Poly(pPDst, pVSrc)  convert roots into polynomial      //
    //------------------------------------------------------------------//
    static void Root2Poly(byte[] abDst, byte[] abSrc)
    {
    int i, j;
        Arrays.fill(abDst, (byte)0);
        abDst[0] = 1;
        for(j = 0; j < NPARITY; j++){
            for(i = j; i >= 0; i--){
                abDst[i+0x1] = GFSub(abDst[i+0x1],
                        GFMpy(abDst[i], abSrc[j]));}}
    }

    //------------------------------------------------------------------//
    //      Poly2Root                                                   //
    //------------------------------------------------------------------//
    static int Poly2Root(byte[] abDst, byte[] abSrc, int n)
    {
    int  i, j;
    byte bLoc;                              // current locator
    byte bSum;                              // current sum
    byte iDst;                              // index to abDst
        if(n == 0)
            return 0;    
        iDst = 0;
        bLoc = 1;
        for(j = 0; j < NCDWRD; j++){
            bSum = 0;                       // sum up terms
            for(i = 0; i <= n; i++){
                bSum = GFMpy(bSum, bLoc);
                bSum = GFAdd(bSum, abSrc[i]);}
            if(bSum == 0){                  // if a root
                if(iDst > n){               //    exit if too many roots
                    return(1);}
                abDst[iDst] = bLoc;         //    append locator
                iDst++;}
            bLoc = GFMpy(bLoc, ALPHA);}     // advance locator
        if(iDst != n)                       // exit if not enough roots
            return(1);
        return(0);                          // indicate success
    }

    //------------------------------------------------------------------//
    //      InitGF  Initialize Galios Stuff                             //
    //------------------------------------------------------------------//
    public static void InitGF()
    {
    int  i, j;
    byte b;
        b = 1;
        for(i = 0; i < 512; i++){       // init abExp[]
            abExp[i] = b;
            b = GFMpy0(b, ALPHA);}
        abLog[0] = (byte)0xff;          // init abLog[]
        for(i = 0; i < 255; i++){
            abLog[(int)abExp[i] & 0xff] = (byte)i;}
        b = FCR;                        // init generator poly roots
        for(i = 0; i < NPARITY; i++){
            abGenRoots[i] = b;
            b = GFMpy(b, ALPHA);}
        Root2Poly(abGenPoly, abGenRoots); // init generator poly
    }

    //------------------------------------------------------------------//
    //      Encode                                                      //
    //------------------------------------------------------------------//
    static void Encode()
    {
    int  i, j;
    byte bQuot;                         // quotient byte
    byte bRem0, bRem1;                  // partial remainders
        Arrays.fill(abParities, (byte)0);     // generate parities
        for(j = 0; j < NDATA; j++){
            bQuot = GFAdd(abCdWrd[j], abParities[0]);
            bRem0 = 0;
            for(i = NPARITY; i != 0;){
                bRem1 = GFSub(bRem0, GFMpy(bQuot, abGenPoly[i]));
                i--;
                bRem0 = abParities[i];
                abParities[i] = bRem1;}}
        for(i = 0; i < NPARITY; i++){   // append parities
            abCdWrd[NDATA+i] = GFSub((byte)0, abParities[i]);}
    }

    //------------------------------------------------------------------//
    //      GenSyndromes    generate standard RS syndromes              //
    //------------------------------------------------------------------//
    static void GenSyndromes()
    {
    int i, j;
        for(j = 0; j < NPARITY; j++){
            abSyndromes[j] = abCdWrd[0];    // generate a syndrome
            for(i = 1; i < NCDWRD; i++){
                abSyndromes[j] = GFAdd(abCdWrd[i],
                    GFMpy(abGenRoots[j], abSyndromes[j]));}}
    }

    //------------------------------------------------------------------//
    //      Euclid      extended Euclid division algorithm              //
    //      generates a series of polynomials:                          //
    //      A[i]S(x) + B[i](x^t) = R[i]                                 //
    //      where the degree of R[i] decreases with each iteration      //
    //      until degree <= MAXERR, then A[i] = Lambda, R[i] = Omega    //
    //      abE0, abE1: left side contains R[]                          //
    //      abE0, abE1: right side contains reversed A[]                //
    //      iE0, iE1: index to end of R[], start of A[]                 //
    //------------------------------------------------------------------//
    static int Euclid()
    {
    int i;
    byte bQuot;                         // quotient
        // E0.R[-1] = x^MAXERR, E0.A[0] = 1
        iE0 = NP1;
        Arrays.fill(abE0, (byte)0);
        abE0[0] = 1;
        abE0[iE0] = 1;
        // E1.R[0] = syndrome polynomial, E1.A[-1] = 0
        iE1 = NP1;
        Arrays.fill(abE1, (byte)0);
        for(i = 1; i < iE1; i++){
            abE1[i] = abSyndromes[NPARITY-i];}
        // abE1[0] = 0;
        // abE1[iE1] = 0;
    
        while(true){                    // while degree of E1.R[] > max error
            while((abE1[0] == 0) &&     //  shift E1 left until E1.R[] msb!=0
                  (iE1 != 0)){          //   or fully shifted left
                iE1--;
                for(i = 0; i < NP1; i++)  
                    abE1[i] = abE1[i+1];
                abE1[NP1] = 0;}
            if(iE1 <= MAXERR){          //  if degree of E1.R[] <= MAXERR, break
                break;}
            while(true){                //  while more divide/multiply sub-steps
                if(abE0[0] != 0){       //   if E0.R[] msb!=0
                    bQuot = GFDiv(abE0[0], abE1[0]);  // Q=E0.R[msb]/E1.R[msb]
                    for(i = 0; i < iE1; i++){         // E0.R[]=E0.R[]-Q*E1.R[]
                        abE0[i] = GFSub(abE0[i], GFMpy(bQuot, abE1[i]));}
                    for(i = iE0; i < NP2; i++){       // E1.A[]=E1.A[]-Q*E0.A[]
                        abE1[i] = GFSub(abE1[i], GFMpy(bQuot, abE0[i]));}}
                if(iE0 == iE1){         //   if sub-steps done, break
                    break;}
                iE0--;                  //   shift E0 left
                for(i = 0; i < NP1; i++)
                    abE0[i] = abE0[i+1];
                abE0[NP1] = 0;}
            abET = abE0;                // swap E0, E1
            abE0 = abE1;
            abE1 = abET;
            iET  = iE0;
            iE0  = iE1;
            iE1  = iET;}
 
        nErr = NP1-iE0;                 // number of errors
        if(iE1 > nErr){                 // if degree E1.R[] too high
            System.out.println("degree E1.R[] too high");
            return 1;}
        while(iE1 < nErr){              // right shift E1 if Omega
            iE1++;                      //  has leading zeroes
            for(i = nErr; i != 0;){
                i--;
                abE1[i+1] = abE1[i];}
            abE1[0] = (byte)0;}
        bQuot = abE0[iE0];              // bQuot = lsb of Lambda
        if(bQuot == 0){
            System.out.println("lsb of Lambda == 0");
            return 1;}
        // LambdaR = E0.A[] / bQuot (without unreversing E0.A[])
        for(i = 0; i <= nErr; i++)
            abLambdaR[i] = GFDiv(abE0[i+iE0], bQuot);
        // Omega = E1.R[] / bquot
        for(i = 0; i <= nErr; i++)
            abOmega[i] = GFDiv(abE1[i], bQuot);
        // Find roots of LambdaR, roots == error locators
        if(0 != Poly2Root(abErrLoc, abLambdaR, nErr)){
            System.out.println("poly2root(LambdaR) failed");
            return 1;}
        return 0;
    }
    
    //----------------------------------------------------------------------//
    //      GenLambda                                                       //
    //----------------------------------------------------------------------//
    static void GenLambda()
    {
    int i, j;
        // Lambda = reverse of LambdaR
        for(i = 0; i <= nErr; i++){
            abLambda[i] = abLambdaR[nErr-i];}
        //      generate DLambda from Lambda  (copy odd terms)
        //          example: derivative of a x^3 + b x^2 + cx + d
        //          (a+a+a) x^2 + (b+b) x + (c)
        //                a x^2 +     0 x +  c
        nDLambda = (nErr+1)/2;
        j = nDLambda-1;
        for(i = nErr-1; i >= 0;){
            abDLambda[j] = abLambda[i];
            j -= 1;
            i -= 2;}
    }

    //------------------------------------------------------------------//
    //      Forney    generate error values using Forney                //
    //------------------------------------------------------------------//
    static int Forney()
    {
    int  i,j;
    byte bDvnd;
    byte bDvsr;
    byte fcorr = (byte)(1 - abLog[FCR&0xff]);   // correction value for FCR
    byte bILoc;                                 // inverse of locator
        Arrays.fill(abForney, (byte)0);
        for(j = 0; j < nErr; j++){
            bDvsr = bDvnd = 0;
            bILoc = GFDiv((byte)1, abErrLoc[j]);
            for(i = nErr; i != 0;){
                i--;
                bDvnd = GFAdd(bDvnd, GFMpy(abOmega[i],
                        GFPow(bILoc, (byte)((nErr-1)-i))));}
            bDvnd = GFMpy(bDvnd, GFPow(abErrLoc[j], fcorr));
            for(i = nDLambda; i != 0;){
                i--;
                bDvsr = GFAdd(bDvsr, GFMpy(abDLambda[i],
                        GFPow(bILoc, GFMpy((byte)2, (byte)(nDLambda-1-i)))));}
            if(bDvsr == 0){
                System.out.println("Forney divide by 0");
                return 1;}
            abForney[j] = GFSub((byte)0, GFDiv(bDvnd, bDvsr));}
        return 0;
    }
    
    //------------------------------------------------------------------//
    //      Decode                                                      //
    //------------------------------------------------------------------//
    static int Decode()
    {
    int i, j;
        GenSyndromes();             // generate syndromes
        if(0 != Euclid())           // Euclid
            return 1;
        GenLambda();                // Forney
        if(0 != Forney())
            return 1;
        for(j = 0; j < nErr; j++){  // fix errors
            i = NCDWRD-1-(abLog[abErrLoc[j]&0xff]&0xff);
            abCdWrd[i] = GFSub(abCdWrd[i], abForney[j]);}
        if(nErr < nErrMin)          // update min and max errors
            nErrMin = nErr;
        if(nErr > nErrMax)
            nErrMax = nErr;
        return 0;
    }

    //------------------------------------------------------------------//
    //  main                                                            //
    //------------------------------------------------------------------//
    public static void main(String[] args) {
    int i, j, k;
        InitGF();
        for(i = 0; i < NDATA; i++)
            abCdWrd[i] = (byte)(i+1);
        Encode();
        for(i = 0; i < NCDWRD; i++)     // copy the codeword
            abCdWrdx[i] = abCdWrd[i];
        Random r = new Random();
        for(k = 0; k < 1024; k++){     // do 1024 tests
            for(j = 0; j < MAXERR; j++){    // create errors
                i = r.nextInt()%NCDWRD;
                if(i < 0)i += NCDWRD;
                abCdWrd[i] ^= (byte)r.nextInt();}
            if(0 != Decode()){              // fix errors
                System.out.println("uncorrectable");
                break;}
            for(i = 0; i < NCDWRD; i++){    // check codeword
                if(abCdWrdx[i] != abCdWrd[i]){
                    break;}}
            if(i != NCDWRD){
                System.out.println("failed");
                break;}
        }
        if(i == NCDWRD){
            System.out.println("passed");
            System.out.println("min errors " + nErrMin +
                    " max errors " + nErrMax);}
    }
}
